/*
** Copyright (C) 1998,1999,2000,2001 Martin Roesch <roesch@clark.net>
** Copyright (C) 1999,2000,2001 Christopher E. Cramer <cec@ee.duke.edu>
** 
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: spp_tcp_stream3.c,v 1.14 2001/06/16 03:51:45 roesch Exp $ */

/* spp_tcp_stream3
 * 
 * Purpose:
 *
 * Construct tcp streams from observed packets
 *
 * Usage:
 *
 * preprocessor stream: <arg set 1>, <arg set 2>, ...
 *
 * Arguments:  one or more of the following separated by commas
 *
 * timeout <timeout value>
 *
 * ports <port 1> ... <port N>
 *
 *
 * <timeout>  - the max time in seconds for which a stream will be 
 *            kept alive if we haven't seen a packet for it
 * <port x>   - a server port to monitor.  we don't want to monitor 
 *            all tcp streams (do we?)
 * <treedepth> - set the max depth of the AVL stream node storage tree.  Value
 *               is the power of 2 that indicates the max depth of the tree
 *               (i.e. 8 = 256 nodes, 10 = 1024 nodes, 16 = 65536 nodes)
 *
 *     example: 
 *       preprocessor stream: timeout 5, ports 21 23 80 8080, treedepth 13
 *
 * Effect:
 *
 * creates a buffer for each observed tcp stream.  upon seeing a CRLF
 * or receiving a maximum number of bytes, generate a packet containing 
 * the reconstructed data
 *
 * Comments:
 *
 * the size of <timeout> will directly impact the amount of memory used.  
 * The longer the timeout the more streams kept in memory, the greater 
 * the memory usage.
 *
 * To Do: 
 *    allow tighter checks on what we monitor - reduce memory usage.
 *
 *
 */
#include <strings.h>
#include <sys/types.h>
#include <sys/time.h>
#include "spp_tcp_stream3.h"
#include <sys/stat.h> /* umask for leak detection */
#include <stdio.h>     /* fopen et al. for leak detection */
#undef HAVE_64

#define SEQ_GT(x,y)     (y - x > 0x7fffffff)
#define SEQ_LT(x,y)     (x - y > 0x7fffffff)
#define SEQ_EQ(x,y)     (x == y)
#define SEQ_GE(x,y)     (x - y < 0x7fffffff)
#define SEQ_LE(x,y)     (y - x < 0x7fffffff)

/* external globals from rules.c */
extern char *file_name;
extern int file_line;
extern int avlnodes;

TcpStream3Data StreamData;
FILE *TcpStream3File;
int sesscount3 = 0;
Packet *dummy;
int pcaplen, hdrsize;
int pktsize;
int pktdatasize;
int flush_size;
int once = 0;

void InitDummyPacket();
void InitNewStream(TcpStream3Session *, Packet *, int, int);
TcpStream3Session *NewStream(int);

/* Leak finder */
int nsession=0;

/*static FILE* leak_finder_report;*/

void 
avlcallback(struct Node* n)
{
    return;
/*       TcpStream3Session* s=n->nsptr;

       fprintf(leak_finder_report, "Session %i, %s\n",
               s->nsession,
               ctime(&(s->itime))
       );*/
}

/*
void
leak_finder_setup(char *dump_filename)
{
       int oumask;

       oumask=umask(022);

       if((leak_finder_report=fopen(dump_filename, "w+"))==NULL){
               fprintf(stderr, "Aiee!  Can't open file \"%s\" for leak detection.  Exiting.\n", dump_filename);
               exit(-1);
       }

       umask(oumask);
}

void
leak_finder(int signum)
{
       time_t c=time(NULL);
       fprintf(leak_finder_report, "Starting dump at: %s\n", ctime(&c));
       fprintf(leak_finder_report, "==========================================\n");
       walkavltree(&avlcallback, rootnode);
       fprintf(leak_finder_report, "==========================================\n");
       fprintf(leak_finder_report, "\n");
       fflush(leak_finder_report);
}
*/
/* End leak finder */

/*
 * Function: SetupTcpStream3()
 *
 * Purpose: Registers the preprocessor keyword and initialization 
 *          function into the preprocessor list.  This is the function that
 *          gets called from InitPreprocessors() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void SetupTcpStream3()
{
    /* link the preprocessor keyword to the init function in 
       the preproc list */
    RegisterPreprocessor("stream3", TcpStream3Init);

    DebugMessage(DEBUG_PLUGIN, "Preprocessor: TcpStream3 is setup...\n");
}


/*
 * Function: TcpStream3Init(u_char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void TcpStream3Init(u_char *args)
{

    DebugMessage(DEBUG_PLUGIN, "Preprocessor: TcpStream3 Initialized\n");

    /* parse the argument list from the rules file */
    ParseTcpStream3Args(args);

    /* Set the preprocessor function into the function list */
    AddFuncToPreprocList(TcpStream3Packet);
    AddFuncToCleanExitList(TcpStream3Exit,NULL);
    AddFuncToRestartList(TcpStream3Restart,NULL);

}


/*
 * Function: ParseTcpStream3Args(char *)
 *
 * Purpose: Process the preprocessor arguements from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function 
 *
 */
void ParseTcpStream3Args(char *args)
{
    char **toks, **secs, **ptoks;
    int num_toks, num_secs, num_ptoks;
    int i,j;

    if(args == NULL)
    {
        FatalError("ERROR %s(%d)=> No arguments to TcpStream3 preprocessor!\n", 
                file_name, file_line);
    }

    StreamData.prunetime = 60;
    StreamData.treedepth = 13;
    StreamData.trunc = 0;

    /* tokenize the argument list */
    secs = mSplit(args, ",", 33, &num_secs, '\\');

    for(i=0;i<num_secs;i++)
    {
        toks = mSplit(secs[i], " ", 33, &num_toks, '\\');

        if(strcmp(toks[0],"timeout") == 0)
        {
            if(num_toks != 2) 
            {
                FatalError("ERROR %s (%d)=> timeout takes 1 argument!\n",
                        file_name,file_line);
            }

            StreamData.prunetime = atoi(toks[1]);

            if(StreamData.prunetime <= 0)
            {
                FatalError("ERROR %s(%d)=> Time between prunings must be "
                        ">0 seconds!\n",file_name,file_line);
            }
        }
        else if(strcmp(toks[0],"treedepth") == 0)
        {
            if(num_toks != 2) 
            {
                FatalError("ERROR %s (%d)=> treedepth takes 1 argument!\n",
                        file_name,file_line);
            }

            StreamData.treedepth = atoi(toks[1]);

            if(StreamData.treedepth <= 0)
            {
                FatalError("ERROR %s(%d)=> tree must have at least one node!\n",
                        file_name,file_line);
            }
        }
        else if(strcmp(toks[0],"truncate") == 0)
        {
            StreamData.trunc = 1;
        }
        else if(strcmp(toks[0],"ports") == 0)
        {
            for(j = 1; j < num_toks; j++)
            {
                ptoks = mSplit(toks[j], "/", 2, &num_ptoks, '\\');

                StreamData.ports[j-1] = atoi(ptoks[0]);
                StreamData.port_collect[j-1] = COLLECT_CLIENT | COLLECT_SERVER;

                if(num_ptoks == 2)
                {
                    if(strcmp(ptoks[1],"c") == 0)
                    {
                        StreamData.port_collect[j-1] = COLLECT_CLIENT;
                    }
                    else if(strcmp(ptoks[1],"s") == 0)
                    {
                        StreamData.port_collect[j-1] = COLLECT_SERVER;
                    }
                    else if(strcmp(ptoks[1],"b") == 0)
                    {
                        StreamData.port_collect[j-1] = COLLECT_SERVER | 
                            COLLECT_CLIENT;
                    }
                    else
                    {
                        FatalError("ERROR %s(%d)=> port modifier must be"
                                " 'c', 's' or 'b'\n",file_name,file_line);
                    }

                    free(ptoks[1]);
                }

                free(ptoks[0]);
#ifdef DEBUG
                DebugMessage(DEBUG_PLUGIN, "New port to monitor: %i/%s%s\n",
                        StreamData.ports[j-1], 
                        StreamData.port_collect[j-1]&COLLECT_SERVER?"s":"",
                        StreamData.port_collect[j-1]&COLLECT_CLIENT?"c":"");
#endif
            }   

            StreamData.num_entries = num_toks-1;
        }
        else
        {
            FatalError("ERROR %s(%d)=> unknown argument to preprocessor"
                    " stream: %s!\n",file_name,file_line,toks[0]);
        }

        for(j=0;j<num_toks;j++)
            free(toks[j]);
                       free(toks);
    }

    for(i=0;i<num_secs;i++)
        free(secs[i]);
    free(secs);
}


/*
 * Function: TcpStream3Packet(Packet *)
 *
 * Purpose:  take a packet, determine if it is tcp.  if so, is it the start of 
 *           a stream?
 *           yes - create new stream.  
 *           no - do we have that stream?  yes - deal w/ packet.
 *                                         no - return
 *
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */
void TcpStream3Packet(Packet *p)
{
    int i;
#ifdef DEBUG
    char sip[16];
    char dip[16];
#endif
    u_long tm;
    TcpStream3Session * sptr = NULL;
    u_long lack, lseq;
    int server_packet;
    u_long make_packet_size;
    int bin;
    u_char CollectInfo;

    if(!once)
    {
        flush_size = pv.max_pattern * 2;
        InitDummyPacket();
        once = 1;
    }
    /* statuses (client-status & server-status) from TCP/IP Illustrated v1, p 241:
       0 - closed
       1 - listen
       2 - SYN_RCVD
       3 - SYN_SENT
       4 - ESTABLISHED  (data xfer mode)
       5 - CLOSE_WAIT
       6 - LAST_ACK
       7 - FIN_WAIT_1
       8 - CLOSING
       9 - FIN_WAIT_2
       10- TIME_WAIT
     */

    if(pv.fake_packet_flag)
    {
        DebugMessage(DEBUG_PLUGIN, 
                "TcpStream3 Preprocessor shouldn't be handling "
                "it's own packets! :-)\n");
        return;
    }

    /* make sure we've got TCP traffic... */
    if(!PacketIsTCP(p) && !p->frag_flag)
    {
        return;
    }

    /* don't accept packets w/ bad checksums */
    if(p->csum_flags & CSE_IP || p->csum_flags & CSE_TCP)
    {
        DebugMessage(DEBUG_PLUGIN, 
                "Discarding packet based on checksum errors!\n");
        return;
    }

    /* check to see if we're interested in the traffic on these ports */
    for(i = 0; i < StreamData.num_entries; i++)
    {
        if((StreamData.ports[i] == p->dp) || (StreamData.ports[i] == p->sp)) 
            break;
    }

    /* didn't find any ports we're interested in */
    if(i == StreamData.num_entries) 
    {
        return;  /* not a monitored port */
    }

    pc.tcp_stream_pkts++;  /* a packet we use */

    CollectInfo = StreamData.port_collect[i];

    /* check to see if this is a server or client packet */
    if(StreamData.ports[i] == p->sp)
    {
        server_packet = 1;
    }
    else
    {
        server_packet = 0;
    }

    /* get the time from the packet data*/
    tm = p->pkth->ts.tv_sec;

    /* if the prune time has expired, prune the session tree */
    if(tm - StreamData.timestamp >= StreamData.prunetime)
    {
        TcpStream3PruneTree(rootnode, p); 
        DebugMessage(DEBUG_PLUGIN, "AVL nodes after pruning: %d\n", avlnodes);
        StreamData.timestamp = tm;
    }

    /* set last ack and seq */
    lack = ntohl(p->tcph->th_ack);
    lseq = ntohl(p->tcph->th_seq);

#ifdef DEBUG
    strncpy(sip, inet_ntoa(p->iph->ip_src), 16);
    strncpy(dip, inet_ntoa(p->iph->ip_dst), 16);
    printf("Packet (%x %lX/%lX %i) - %s:%i -> %s:%i (%i)\n",p->tcph->th_flags,
            lseq,lack,p->tcph->th_win,sip,p->sp,dip,p->dp,p->dsize);
#endif 

    /* get the stream pointer for this session; bin will have key*/
    sptr = TcpStream3CheckSession(p,StreamData.ports[i],server_packet,&bin);

    /* STATE MACHINE */
    /* if none exist, start a new one */
    if(sptr == (TcpStream3Session *)NULL)
    {
        /* if this is a SYN packet, start a new stream */
        if(p->tcph->th_flags == R_SYN)
        { 
            sptr = NewStream(bin);

            if(sptr != NULL)
                InitNewStream(sptr, p, lseq, lack);
        }

        return; /* either way, we are done here */
    }

    /* update the last access time for this stream */
    sptr->timestamp = tm;

    /* set any window size variations */
    if(server_packet)
    {
        sptr->s_win = ntohs(p->tcph->th_win);
    }
    else
    {
        sptr->c_win = ntohs(p->tcph->th_win);
    }

    /* check for people trying to establish another stream with the same
     * ports and IPs
     */ 
    if(p->tcph->th_flags == R_SYN)     /* but we already have a stream */
    {
        return;                 
    }

    /* STATE MACHINE */
    /* check for session closing down */
    if(p->tcph->th_flags & R_RST)
    {
        /* check RST seq - possible fake RST to confuse NIDS */
        if(!server_packet && (SEQ_LT(lseq,sptr->s_last_ack) || 
                    SEQ_GE(lseq, (u_long)sptr->s_win + sptr->s_last_ack)) )
        {
            DebugMessage(DEBUG_PLUGIN, "Ignoring bad RST from server,"
                    " seq: %lX not in range"
                    " [%lX and %lX)!\n",lseq,sptr->s_last_ack,
                    sptr->s_last_ack+(u_long)sptr->s_win);

            /* NOTE: This should probably be flagged as possible evasion */
            return;
        }

        if(server_packet && (SEQ_LT(lseq, sptr->c_last_ack) || 
                    SEQ_GE(lseq, (u_long)sptr->c_win + sptr->c_last_ack)) )
        {
            DebugMessage(DEBUG_PLUGIN, "Ignoring bad RST from client,"
                    " seq: %lX not in range"
                    " [%lX and %lX)!\n",lseq,sptr->c_last_ack,
                    sptr->c_last_ack+(u_long)sptr->c_win);

            /* NOTE: This should probably be flagged as possible evasion */
            return;
        }

        /* handle RST by transitioning to the CLOSED state */
        sptr->client_status = CLOSED;
        sptr->server_status = CLOSED;
        sptr->c_last_ack = lseq + 1;
        sptr->timestamp = tm;

        DebugMessage(DEBUG_PLUGIN, "flushing reset stream\n");
        /* clear the packets from the stream tree and prune */
        TcpStream3PruneSession(bin, sptr);

        return;
    }

    /* STATE MACHINE */
    /* look for part 2 of the TWH */
    if(p->tcph->th_flags == (R_SYN | R_ACK))
    { 
        /* SYN/ACK */
        if(!server_packet)
        { 
            /* NOTE: This should probably be flagged as possible evasion */
            return;  /* Client should not SYN/ACK */
        }

        if(lack != sptr->c_first_seq)
        {
            return; /* False SYN/ACK - bad seq*/
        }

        if((sptr->client_status == SYN_SENT) && 
                (sptr->server_status == SYN_RCVD))
        {
            sptr->s_first_seq = lseq+1;
            sptr->s_last_ack = lseq+1; 

            /* client happy, should send ACK */
            sptr->client_status = ESTABLISHED; 
        }

        return;
    }

    DebugMessage(DEBUG_PLUGIN, "status: %i/%i\n",sptr->server_status, 
            sptr->client_status);

    /* STATE MACHINE */
    /* Store Data if we see data and are ESTABLISHED or in CLOSE_WAIT
       and we are collecting data from this half of the stream
     */
    if((sptr->client_status == ESTABLISHED || 
            sptr->client_status == CLOSE_WAIT) && 
            (CollectInfo & COLLECT_CLIENT)      && 
            !server_packet)
    {
        /* copy data and then truncate if desired */
        TcpStream3StoreData(sptr, p, server_packet);

        if(StreamData.trunc)
        {
            p->dsize = 0;
        }
#ifdef DEBUG
        TcpStream3PrintDataList(sptr, !server_packet);
#endif
    }

    /* STATE MACHINE */
    if((sptr->server_status == ESTABLISHED || 
            sptr->server_status == CLOSE_WAIT) && 
            (CollectInfo & COLLECT_SERVER) &&
            server_packet)
    {
        /* copy data and then truncate if desired */
        TcpStream3StoreData(sptr, p, server_packet);

        if(StreamData.trunc)
        {
            p->dsize = 0;
        }
#ifdef DEBUG
        TcpStream3PrintDataList(sptr, !server_packet);
#endif
    }

    /* Check here to see if we inject new server packet */
    if(!server_packet && 
            sptr->server_status >= ESTABLISHED && 
            lack > sptr->s_last_ack)
    { 
        make_packet_size = TcpStream3CheckAck(sptr, lack, server_packet);

        if (make_packet_size > 0) 
        {
            if(make_packet_size > pktdatasize)
            {
                LogMessage("Oops, bad packet length from TcpStream3AckCheck()\n");
                make_packet_size = pktdatasize;
            }
#ifdef DEBUG
            TcpStream3PrintDataList(sptr, !server_packet);
#endif
            TcpStream3Packetize(sptr, p, make_packet_size, server_packet);

        }
    }

    /* Check here to see if we inject client new packet */
    if(server_packet && 
            sptr->client_status >= ESTABLISHED && 
            lack > sptr->c_last_ack)
    {  
        make_packet_size = TcpStream3CheckAck(sptr, lack, server_packet);

        if (make_packet_size > 0) 
        {
#ifdef DEBUG
            TcpStream3PrintDataList(sptr, !server_packet);
#endif
            TcpStream3Packetize(sptr, p, make_packet_size, server_packet);

        }
    }

    /* if something's been acknowledged, update last_acked 
     */
    if(p->tcph->th_flags & R_ACK)
    {
        if(server_packet)
        {
            sptr->c_last_ack = lack;
            if(sptr->server_status == FIN_WAIT_1)
            {
                sptr->server_status = CLOSING;
            }
            else if(sptr->client_status == CLOSING)
            {
                sptr->client_status = TIME_WAIT;
            }
        }

        if(!server_packet)
        {
            sptr->s_last_ack = lack;

            if(sptr->client_status == FIN_WAIT_1)
            {
                sptr->client_status = CLOSING;
            }
            else if(sptr->client_status == CLOSING)
            {
                sptr->client_status = TIME_WAIT;
            }
        }
    }

    /* STATE MACHINE */
    if((sptr->client_status == ESTABLISHED) && 
            (sptr->server_status == SYN_RCVD))
    {
        if(server_packet) 
            return; /* shouldn't see a server packet here */

        if(lack != sptr->s_first_seq) 
            return; /* False ACK - bad seq */

        if(p->tcph->th_flags == R_ACK) sptr->server_status = ESTABLISHED;
        return;
    }

    /* STATE MACHINE */
    if((sptr->server_status == ESTABLISHED) && 
            (sptr->client_status == ESTABLISHED) && 
            (p->tcph->th_flags & R_FIN))
    {
        if(server_packet)
        {
            sptr->server_status = FIN_WAIT_1;  /* server does active close */
            sptr->client_status = CLOSE_WAIT;
        }
        else
        {
            sptr->client_status = FIN_WAIT_1; /* client does active close */
            sptr->server_status = CLOSE_WAIT;
        }
    }

    /* STATE MACHINE */
    if(sptr->server_status == CLOSE_WAIT && 
            server_packet                && 
            p->tcph->th_flags & R_FIN)
    {
        sptr->server_status = LAST_ACK;
        sptr->client_status = TIME_WAIT;
    }

    /* STATE MACHINE */
    if(sptr->client_status == CLOSE_WAIT && 
            !server_packet               && 
            p->tcph->th_flags & R_FIN)
    {
        sptr->server_status = TIME_WAIT;
        sptr->client_status = LAST_ACK;
    }

    if((sptr->server_status == LAST_ACK || sptr->server_status == TIME_WAIT) &&
       (sptr->client_status == LAST_ACK || sptr->client_status == TIME_WAIT))
    {
        DebugMessage(DEBUG_PLUGIN, "flushing closed stream\n");
        TcpStream3PruneSession(bin, sptr);
    }     

    return;
}


/************************************************************************
 ************************************************************************
 ************************************************************************/
u_long TcpStream3DataSize(TcpStream3Session *sptr, int server)
{
    TcpStream3PacketData *pdata;
    u_long size = 0;

    DebugMessage(DEBUG_PLUGIN, "Finding maximum amount of data in buffers\n");

    if(server)
    {
        size = sptr->s_stream_size;
        pdata = sptr->s_data;
    }
    else
    {
        size = sptr->c_stream_size;
        pdata = sptr->c_data;
    }

    /* check to make sure we have some data */
    if(pdata == (TcpStream3PacketData *)NULL)
    {
        return 0;
    }

    return size;
}


/************************************************************************
 ************************************************************************
 ************************************************************************/
u_long TcpStream3CheckAck(TcpStream3Session *sptr, u_long lack, 
                          int server_packet)
{
    u_long first_seq;
    u_long size = 0;

    DebugMessage(DEBUG_PLUGIN, "Finding length of packet based on ACK\n");

    if(!server_packet)  /* client ack */
    {
        /* no data in the stream yet */
        if(sptr->s_data == (TcpStream3PacketData *)NULL)
        {
            return 0;
        }

        first_seq = sptr->s_data->seq;

        if( SEQ_GE(first_seq, lack) ) return 0;

        /* if total size of data >= pv.max_pattern * 2 then send packets */
        size = sptr->s_stream_size;

        if (size >= flush_size)
        {
            DebugMessage(DEBUG_PLUGIN, "exceeded flush size\n");
            return size;
        }
        else
        {
            return 0;
        }
    }
    else
    {
        if(sptr->c_data == (TcpStream3PacketData *)NULL)
        {
            return 0;
        }    	

        first_seq = sptr->c_data->seq;

        if( SEQ_GE(first_seq, lack) ) return 0;

        /* if total size of data >= pv.max_pattern * 2 then send packets */
        size = sptr->c_stream_size;

        if (size >= flush_size)
        {
            DebugMessage(DEBUG_PLUGIN, "exceeded flush size\n");
            return size;
        }
        else
        {
            return 0;
        }
    }
}    


/************************************************************************
 ************************************************************************
 ************************************************************************/
void TcpStream3StoreData( TcpStream3Session * sptr, Packet * p, int server_packet)
{
    /* do we need a +1 on seq if SYN or FIN? */
    TcpStream3PacketData *pdata, *cur, *tmp;
    u_long seq;

    DebugMessage(DEBUG_PLUGIN, "pulling data from packet\n");

    if(p->dsize == 0)
    {
        DebugMessage(DEBUG_PLUGIN, "No data in packet\n");
        return;
    }

    seq = ntohl(p->tcph->th_seq);

    /* need to see if we should deal w/ this data */
    if(!server_packet && SEQ_LT(seq, sptr->c_last_ack) )
    {

        DebugMessage(DEBUG_PLUGIN, "Seeing client retransmissions, seq: %lX, "
                     "c_last_ack: %lX\n",seq, sptr->c_last_ack);
        return;
    }
    else if(server_packet && SEQ_LT(seq, sptr->s_last_ack) )
    {
        DebugMessage(DEBUG_PLUGIN, "Seeing server retransmissions, seq: %lX, "
                     "s_last_ack: %lX\n",seq, sptr->s_last_ack);
        return;
    }
    else if(!server_packet && seq + 
            (u_long)p->dsize - sptr->c_last_ack > sptr->s_win)
    {
        DebugMessage(DEBUG_PLUGIN, "Client data out of window\n");
        return;
    }
    else if(server_packet && seq + 
            (u_long)p->dsize - sptr->s_last_ack > sptr->c_win)
    {
        DebugMessage(DEBUG_PLUGIN, "Server data out of window\n");
        return;
    }

    /* pdata = (TcpStream3PacketData *) calloc(1,sizeof(TcpStream3PacketData)); */
    pdata = (TcpStream3PacketData *) malloc(sizeof(TcpStream3PacketData)); 
    DebugMessage(DEBUG_PLUGIN, "Location of new pdata: %x\n",pdata);

    if(!pdata)
    {
        FatalError("ERROR: Unable to allocate memory for stream node!\n");
    }

    pdata->seq = seq;
    pdata->size = p->dsize;
    pdata->next = pdata->prev = (TcpStream3PacketData *)NULL;

    /* pdata->data = (u_char *)calloc(pdata->size, sizeof(u_char));*/
    pdata->data = (u_char *)malloc(pdata->size * sizeof(u_char));
    DebugMessage(DEBUG_PLUGIN, "Location of new pdata->data: %x\n",pdata->data);

    if(!pdata->data)
    {
        FatalError("ERROR: Unable to allocate memory for stream node!\n");
    }

    memcpy(pdata->data, p->data, pdata->size);

    if(server_packet)
    {
        if(sptr->s_data == NULL)
        {
            DebugMessage(DEBUG_PLUGIN, "initializing sptr server data list "
                    "(this one works)\n");
            sptr->s_data = pdata;
            sptr->s_data_end = pdata;
            sptr->s_stream_size += pdata->size;
            sptr->s_count++;
            DebugMessage(DEBUG_PLUGIN, "Adding %d bytes to server stream\n", 
                    pdata->size);
            return;
        }
        else
        {
            cur = sptr->s_data_end;

            /* try to put it at the end of the list first, that's the most
             * likely place where this segment will go 
             *
             * If the sequence number of the last node on the list is less than
             * the seq number of the new data put the new node at the end 
             * of the list
             */ 
            DebugMessage(DEBUG_PLUGIN, "checking end node\n");

            if(SEQ_GE(pdata->seq, cur->seq))
            {
                DebugMessage(DEBUG_PLUGIN, "appending stream segment\n");
                cur->next = pdata;  /* set the next node to the new data */

                pdata->prev = cur;  /* set the new node's prev ptr to point
                                     * to the former last node */  

                sptr->s_data_end = pdata; /* set the list end pointer to the 
                                           * new node */

                sptr->s_count++;
                sptr->s_stream_size += pdata->size; /* add the new node to the
                                                     * stream size */  
                DebugMessage(DEBUG_PLUGIN, 
                        "Adding %d bytes to server stream\n", 
                        pdata->size);
                return;
            }

            /* no dice, do an insertion sort */
            cur = sptr->s_data;

            /* check if the current packet comes before the first packet in the list */
            if(SEQ_LT(pdata->seq, cur->seq))
            {
                DebugMessage(DEBUG_PLUGIN, "inserting out of order segment\n");
                pdata->next = cur;
                cur->prev = pdata;
                sptr->s_stream_size += pdata->size;
                sptr->s_count++;
                DebugMessage(DEBUG_PLUGIN, 
                        "Adding %d bytes to server stream\n", 
                        pdata->size);
                return;
            }

            DebugMessage(DEBUG_PLUGIN, "performing insertion sort\n");
            while(cur != NULL)
            {
                if(SEQ_GE(pdata->seq, cur->seq))
                {
                    DebugMessage(DEBUG_PLUGIN, "stream segment inserted\n");
                    tmp = cur->next;
                    cur->next = pdata;
                    pdata->prev = cur;
                    pdata->next = tmp;
                    sptr->s_data_end = pdata;
                    sptr->s_stream_size += pdata->size;
                    sptr->s_count++;
                    DebugMessage(DEBUG_PLUGIN, 
                            "Adding %d bytes to server stream\n", 
                            pdata->size);
                    return;
                }

                cur = cur->next;
            }

            /* if we get to here things are seriously fucked up */
            printf("That's bad, m'kay (server side)\n");
            printf("pdata->seq = %d, sptr->s_data_end->seq = %p sptr->s_data->seq = %d\n", pdata->seq, sptr->s_data_end, sptr->s_data->seq);
            printf("freeing unplaced segment node data\n");
            free(pdata->data);
            printf("freeing unplaced segment node\n");
            free(pdata);
            return;
        }
    }
    else
    {
        if(sptr->c_data == NULL)
        {
            sptr->c_data = pdata;
            sptr->c_data_end = pdata;
            sptr->c_stream_size += pdata->size;
            sptr->c_count++;
            DebugMessage(DEBUG_PLUGIN, 
                    "Adding %d bytes to client stream\n", 
                    pdata->size);
            return;
        }
        else
        {
            cur = sptr->c_data_end;

            /* try to put it at the end of the list first, that's the most
             * likely place where this segment will go 
             *
             * If the sequence number of the last node on the list is less than
             * the seq number of the new data put the new node at the end 
             * of the list
             */ 
            if(SEQ_GE(pdata->seq, cur->seq))
            {
                cur->next = pdata;  /* set the next node to the new data */

                pdata->prev = cur;  /* set the new node's prev ptr to point
                                     * to the former last node */  

                sptr->c_data_end = pdata; /* set the list end pointer to the 
                                           * new node */

                sptr->c_count++;
                sptr->c_stream_size += pdata->size; /* add the new node to the
                                                     * stream size */  
                DebugMessage(DEBUG_PLUGIN, 
                        "Adding %d bytes to server stream\n", 
                        pdata->size);
                return;
            }

            /* no dice, do an insertion sort */
            cur = sptr->c_data;

            /* check if the current packet comes before the first packet in the list */
            if(SEQ_LT(pdata->seq, cur->seq))
            {
                DebugMessage(DEBUG_PLUGIN, "inserting out of order segment\n");
                pdata->next = cur;
                cur->prev = pdata;
                sptr->c_stream_size += pdata->size;
                sptr->c_count++;
                DebugMessage(DEBUG_PLUGIN, 
                        "Adding %d bytes to server stream\n", 
                        pdata->size);
                return;
            }


            while(cur != NULL)
            {
                if(SEQ_GE(pdata->seq, cur->seq))
                {
                    tmp = cur->next;
                    cur->next = pdata;
                    pdata->prev = cur;
                    pdata->next = tmp;
                    sptr->c_data_end = pdata;
                    sptr->c_count++;
                    sptr->c_stream_size += pdata->size;
                    DebugMessage(DEBUG_PLUGIN, 
                            "Adding %d bytes to server stream\n", 
                            pdata->size);
                    return;
                }

                cur = cur->next;
            }

            /* if we get to here things are seriously fucked up */
            printf("That's bad, m'kay (client size)\n");
            printf("pdata->seq = %d, sptr->s_data_end->seq = %p sptr->s_data->seq = %d\n", pdata->seq, sptr->s_data_end, sptr->s_data->seq);
            printf("freeing unplaced segment node data\n");
            free(pdata->data);
            printf("freeing unplaced segment node\n");
            free(pdata);

            return;

        }
    }
}

/*
 * Function: TcpStream3Time()
 *
 * Purpose:  what can i say, i like having my own timing routines
 *
 * Arguments: none
 *
 * Returns: u_long number of seconds since beginning of time (unix-centric)
 *
 */
u_long TcpStream3Time(Packet *p)
{
    struct timeval t;

    if(p != NULL)
    {
        return(p->pkth->ts.tv_sec);
    }
    else
    {
        gettimeofday(&t, (struct timezone *)NULL);
        return(t.tv_sec);
    }
}

/*
 * Function: TcpStream3CheckSession(Packet *, int sp, int server)
 *
 * Purpose:  return the TcpStream3Session of current packet (if any) w/ 
 * respect to server port sp
 *
 *
 * Arguments: 
 *
 * Returns: appropriate TcpStream3Session (or null)
 *
 */
TcpStream3Session * TcpStream3CheckSession(Packet *p, int sp, int server, int *bin)
{
    struct Node *nptr;
    u_int32_t tmp;
    /*char tmp[32];*/

    /* turn the high port/lo port into a 32-bit hashing key */
    if(p->sp > p->dp)
    {
        tmp = (p->sp << 16) | p->dp;
    }
    else
    {
        tmp = (p->dp << 16) | p->sp;
    }

    *bin = (u_int32_t)p->iph->ip_src.s_addr ^ 
        (u_int32_t)p->iph->ip_dst.s_addr ^ 
        tmp;

    DebugMessage(DEBUG_PLUGIN, "HASH: %x\n", *bin);

    nptr = findnode(rootnode, *bin);

    if (nptr == NULL)
        return((TcpStream3Session *)NULL);
    else
        return(nptr->nsptr);

}



/*
 * Function: TcpStream3PruneTree()
 *
 * Purpose:  loop through avl tree and prune sessions
 *
 * Arguments: none
 *
 * Returns: void function
 *
 */

void TcpStream3PruneTree(struct Node *nodeptr, Packet *p)
{
    DebugMessage(DEBUG_PLUGIN, "AVL Nodes: %d\n", avlnodes);

    if (nodeptr == NULL)
        return;

    TcpStream3PruneTree(nodeptr->left, p);

    if((TcpStream3Time(p) - nodeptr->nsptr->timestamp) >= StreamData.prunetime)
    {
        TcpStream3PruneSession(nodeptr->data, nodeptr->nsptr);
    }

    TcpStream3PruneTree(nodeptr->right, p);
}


/*
 * Function: TcpStream3PruneSession()
 *
 * Purpose:  kill all inactive sessions
 *
 * Arguments: none
 *
 * Returns: void function
 *
 */
void TcpStream3PruneSession(int key, TcpStream3Session * sptr)
{
    struct pseudoheader ph;    /* pseudo header declaration */
    grinder_t tmp_grinder;
    int psize;

#ifdef DEBUG
    int prunecount = 0;
#endif

    if(sptr == NULL)
    {
        LogMessage("Got NULL sptr, please tell Marty! (that's bad, m'kay?)\n");
        return;
    }

    tmp_grinder = grinder;
    grinder = DecodeNullPkt;
    pv.fake_packet_flag = 1;

    psize = TcpStream3DataSize(sptr,1);

#ifdef DEBUG
    if(psize > sptr->s_stream_size)
    {
        printf(" [!] stream size (%d) != psize (%d)\n", 
                sptr->s_stream_size, psize);
    }
#endif

    if(psize > 0) 
    {
        pc.rebuilt_tcp ++;  /* new phoney packet */
        dummy->pkth->caplen = dummy->pkth->len = hdrsize + psize;
        dummy->iph->ip_src.s_addr = sptr->sip;
        dummy->iph->ip_dst.s_addr = sptr->cip;
        dummy->iph->ip_len = htons(40 + psize);
        dummy->iph->ip_csum = checksum((u_short *)dummy->iph, 20, (u_short *)NULL, 0);
        dummy->tcph->th_sport = sptr->sp;
        dummy->tcph->th_dport = sptr->cp;
        ph.sip = (u_int32_t)(dummy->iph->ip_src.s_addr);
        ph.dip = (u_int32_t)(dummy->iph->ip_dst.s_addr);
        ph.zero = 0;
        ph.protocol = dummy->iph->ip_proto;
        ph.tcplen = htons(20 + psize);
        dummy->tcph->th_seq = TcpStream3FillBuffer(sptr, (u_char *)dummy->pkth+hdrsize+pcaplen, psize, 1);
        dummy->tcph->th_ack = sptr->c_last_ack;
        dummy->tcph->th_win = htons(sptr->s_win);
        dummy->tcph->th_sum = checksum((u_int16_t *)&ph, 12, (u_int16_t *)(dummy->tcph), 20 + psize);
        ProcessPacket(NULL, dummy->pkth, dummy->pkt);

#ifdef DEBUG
        prunecount++;
#endif
    }

    psize = TcpStream3DataSize(sptr,0);

#ifdef DEBUG
    if(psize > sptr->c_stream_size)
    {
        printf(" [!] stream size (%d) != psize (%d)\n", 
                sptr->c_stream_size, psize);
    }
#endif
    if(psize > 0)
    {
        pc.rebuilt_tcp ++;  /* new phoney packet */
        dummy->pkth->caplen = dummy->pkth->len = hdrsize + psize;
        dummy->iph->ip_src.s_addr = sptr->cip;
        dummy->iph->ip_dst.s_addr = sptr->sip;
        dummy->iph->ip_len = htons(40 + psize);
        dummy->iph->ip_csum = checksum((u_short *)dummy->iph, 20, (u_short *)NULL, 0);
        dummy->tcph->th_sport = sptr->cp;
        dummy->tcph->th_dport = sptr->sp;
        ph.sip = (u_int32_t)(dummy->iph->ip_src.s_addr);
        ph.dip = (u_int32_t)(dummy->iph->ip_dst.s_addr);
        ph.zero = 0;
        ph.protocol = dummy->iph->ip_proto;
        ph.tcplen = htons(20 + psize);
        dummy->tcph->th_seq = TcpStream3FillBuffer(sptr, (u_char *)dummy->pkth+hdrsize+pcaplen, psize, 0);
        dummy->tcph->th_ack = sptr->s_last_ack;
        dummy->tcph->th_win = htons(sptr->c_win);
        dummy->tcph->th_sum = checksum((u_int16_t *)&ph, 12, (u_int16_t *)(dummy->tcph), 20 + psize);
        ProcessPacket(NULL, dummy->pkth, dummy->pkt);

#ifdef DEBUG
        prunecount++;
#endif
    }

    DebugMessage(DEBUG_PLUGIN, "freeing stream node\n");
/*
    free(sptr);
*/
    deletenode(key, &rootnode, &rootnode);


#ifdef DEBUG
    nodecount = 0;
    countnodes(rootnode); 
    DebugMessage(DEBUG_PLUGIN, "Done PRUNING!  %d deleted, %d remain\n",prunecount,nodecount);

#endif

    grinder = tmp_grinder;
    pv.fake_packet_flag = 0;
    return;
}


/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
void TcpStream3Packetize(TcpStream3Session *sptr, Packet *pb, u_long psize, int server_packet)
{
    struct pseudoheader ph;    /* pseudo header declaration */
    grinder_t tmp_grinder;

#ifdef DEBUG
    printf("Trying to make a new packet!  psize: %lu\n", psize);
#endif

    pc.rebuilt_tcp ++;  /* new phoney packet */

    dummy->pkth->ts.tv_sec = pb->pkth->ts.tv_sec;
    dummy->pkth->ts.tv_usec = pb->pkth->ts.tv_usec;
    dummy->pkth->caplen = dummy->pkth->len = hdrsize + psize;
    dummy->iph->ip_src.s_addr = pb->iph->ip_dst.s_addr;
    dummy->iph->ip_dst.s_addr = pb->iph->ip_src.s_addr;
    dummy->iph->ip_len = htons(40 + psize);
    dummy->iph->ip_csum = checksum((u_short *)dummy->iph, 20, (u_short *)NULL, 0);
    dummy->tcph->th_sport = pb->tcph->th_dport;
    dummy->tcph->th_dport = pb->tcph->th_sport;
    ph.sip = (u_int32_t)(pb->iph->ip_dst.s_addr);
    ph.dip = (u_int32_t)(dummy->iph->ip_dst.s_addr);
    ph.zero = 0;
    ph.protocol = dummy->iph->ip_proto;
    ph.tcplen = htons(20 + psize);

    dummy->tcph->th_seq = TcpStream3FillBuffer(sptr, (u_char *)dummy->pkth+hdrsize + pcaplen, psize, !server_packet);

    if(!server_packet)
    {
        dummy->tcph->th_ack = sptr->s_last_ack;
        dummy->tcph->th_win = htons(sptr->c_win);
    }
    else
    {
        dummy->tcph->th_ack = sptr->c_last_ack;
        dummy->tcph->th_win = htons(sptr->s_win);
    }

    dummy->tcph->th_sum = checksum((u_int16_t *)&ph, 12, (u_int16_t *)(dummy->tcph), 20 + psize);

    tmp_grinder = grinder;
    grinder = DecodeNullPkt;
    pv.fake_packet_flag = 1;
    ProcessPacket(NULL, dummy->pkth, dummy->pkt);
    pv.fake_packet_flag = 0;
    grinder = tmp_grinder;
}


/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
u_long TcpStream3FillBuffer(TcpStream3Session *sptr, u_char * buf,
                            u_long psize, int server)
{
    TcpStream3PacketData *pdata, *tmp;
    u_long first_seq;
    u_long index = 0;
    int ptrnull = 0;

#ifdef DEBUG
    printf("In fill buffer on %s side with size: %lu\n",server?"server":"client",psize);

    TcpStream3PrintDataList(sptr, server);
#endif

    /* we really need to 0 out the buffer in case we have gaps */
    DebugMessage(DEBUG_PLUGIN, "Clearing buffer\n");
    memset(buf,0,pktdatasize);

    if(server)
    {
        DebugMessage(DEBUG_PLUGIN, "DataNodes before buffer fill: %d\n", sptr->s_count);
        pdata = sptr->s_data;
        sptr->s_stream_size = 0;
    }
    else
    {
        DebugMessage(DEBUG_PLUGIN, "DataNodes before buffer fill: %d\n", sptr->c_count);
        pdata = sptr->c_data;
        sptr->c_stream_size = 0;
    }

    first_seq = pdata->seq;

    if(pdata == (TcpStream3PacketData *)NULL)
    {
        ptrnull = 1;
    }

    while(pdata != (TcpStream3PacketData *)NULL)
    {
        index = pdata->seq - first_seq;

        if(index < pktdatasize)
        {
            DebugMessage(DEBUG_PLUGIN, "Copying data into buffer\n");
            memcpy(buf + index, pdata->data, pdata->size);
        }

        DebugMessage(DEBUG_PLUGIN, "pdata loc: %x\n",pdata);
        tmp = pdata;
        pdata = pdata->next;

        DebugMessage(DEBUG_PLUGIN, "tmp loc: %x\n",tmp);
        DebugMessage(DEBUG_PLUGIN, "tmp->data loc: %x\n",tmp->data);
        if(tmp->data != (u_char *)NULL)
        {
            DebugMessage(DEBUG_PLUGIN, "freeing segment data\n");
            free(tmp->data);
        }

        DebugMessage(DEBUG_PLUGIN, "freeing segment data node\n");
        free(tmp);

        if(server)
            sptr->s_count--;
        else
            sptr->c_count--;
    }

    if(server)
    {
        DebugMessage(DEBUG_PLUGIN, "After fill buffer PacketData count: %d\n", sptr->s_count);
        sptr->s_data = (TcpStream3PacketData *)NULL;
    }
    else
    {
        DebugMessage(DEBUG_PLUGIN, "After fill buffer PacketData count: %d\n", sptr->c_count);
        sptr->c_data = (TcpStream3PacketData *)NULL;
    }

    return first_seq;
}



/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
void TcpStream3Exit(int signal, void *rg)
{
  
    StreamData.prunetime = -1;
    TcpStream3PruneTree(rootnode, NULL);

}

void TcpStream3Restart(int signal, void *arg)
{

}

void TcpStream3PrintDataList(TcpStream3Session *sptr, int server)
{
    TcpStream3PacketData *cur;
    int i,j;
    
    if(server)
    {
    	cur = sptr->s_data;
    }
    else
    {
    	cur = sptr->c_data;
    }
    
    printf("Data list for: %s\n",server?"server":"client");

    for (i=0 ; cur != (TcpStream3PacketData *)NULL; cur = cur->next, i++)
    {
    	printf("  %i: %lX(%lu) - \"",i,(u_long)cur->seq,(u_long)cur->size);
        for (j=0;j<(cur->size<10?cur->size:10);j++)
            printf("%x ",cur->data[i]);
        printf("\"\n");
    }

    printf("\n");
}



int IsWatchingPort(int hsp, int lsp, int hdp, int ldp)
{
    int port;
    int i;
    
    for(i = 0; i < StreamData.num_entries; i++)
    {
        port = StreamData.ports[i];

        if(((port <= hsp)&&(port>=lsp)) || ((port <= hdp)&&(port >= ldp))) 
        {
            DebugMessage(DEBUG_CONFIGRULES, 
                         "[*] Interested in port %d\n", port);
            return 1;
        }
    }

    return 0;
}




void InitDummyPacket()
{
    /* preallocate a dummy packet */
    dummy = (Packet *) calloc(sizeof(Packet), sizeof(char));

    if(dummy == NULL) 
    {
        return;
    }

    pcaplen = 8 + sizeof(struct timeval);

    pktsize = pcaplen+4+40+flush_size+65535; 
    pktdatasize = 65535;

    /* pcaplen + pad + ip header + tcp header + payload space */
    dummy->pkth = (struct pcap_pkthdr *)calloc(pktsize, sizeof(char)); 

    /* if we had a bad allocation, drop out */
    if(dummy->pkth == NULL)
    {
        return;
    }

    /* preset the dummy packet headers */
    pv.fake_packet_flag = 0;

    dummy->iph = (IPHdr *)((u_char*)dummy->pkth + pcaplen + 4);
    hdrsize = 40 + NULL_HDRLEN;
    dummy->tcph = (TCPHdr *)((u_char *) dummy->iph + 20);
    dummy->pkt = ((u_char*)dummy->pkth + pcaplen);
    dummy->iph->ip_ver = 0x04;
    dummy->iph->ip_hlen = 0x05;
    dummy->iph->ip_proto = IPPROTO_TCP;
    dummy->iph->ip_id = 0;
    dummy->iph->ip_off = htons(0);
    dummy->iph->ip_ttl = 255;
    dummy->iph->ip_csum = 0;
    dummy->iph->ip_tos = 0x10;
    memset(dummy->tcph,0,sizeof(TCPHdr));
    dummy->tcph->th_flags = R_PSH | R_ACK;
    dummy->tcph->th_off = 0x05;

}



TcpStream3Session *NewStream(int bin)
{
    struct Node *nptr;

    inserttree(bin, &rootnode,StreamData.treedepth);

    nptr = findnode(rootnode, bin);

    if(nptr == NULL)
        return NULL;

    nptr->nsptr = (TcpStream3Session *) calloc(sizeof(TcpStream3Session), 
                                               sizeof(char));

    if(nptr->nsptr == NULL)
    {
        FatalError("Unable to allocate new stream node, reduce your treedepth "
                   "and try again\n");
    }

        nptr->nsptr->itime=time(NULL);
        nptr->nsptr->nsession=nsession++;

    if(nptr == NULL)
        return NULL;
    else
        return(nptr->nsptr);
}


void InitNewStream(TcpStream3Session *sptr, Packet *p, int lseq, int lack)
{
    pc.tcp_streams ++;  /* new tcp stream */

    sptr->client_status = SYN_SENT;
    sptr->server_status = SYN_RCVD;
    sptr->s_stream_size = 0;
    sptr->c_stream_size = 0;
    sptr->cip = p->iph->ip_src.s_addr;
    sptr->cp = p->tcph->th_sport;
    sptr->sip = p->iph->ip_dst.s_addr;
    sptr->sp = p->tcph->th_dport;
    sptr->c_first_seq = lseq + 1;
    sptr->c_last_ack = lseq + 1;
    
    if(p->dsize > 0)
    {
        TcpStream3StoreData(sptr, p, 0);
    }
    else
    {
        sptr->c_data = 0;
    }
    
    sptr->s_data = (TcpStream3PacketData *)NULL;
    sptr->timestamp = p->pkth->ts.tv_sec;

    return;
}
